package JointFeldmanProtocol;

import FeldmanVerifiableSecretSharing.ShamirSecretSharing.Polynomial;
import FeldmanVerifiableSecretSharing.VerifiableSecretSharing;
import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import org.bouncycastle.util.Arrays;
import meerkat.protobuf.DKGMessages.*;

import java.math.BigInteger;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

/**
 * Created by Tzlil on 2/5/2016.
 *
 * an implementation of a version of Pedersen's distributed key generation protocol
 */
public class DKG extends VerifiableSecretSharing implements Runnable{

    private Network.User user;
    private Network.MailHandler handler;
    private Polynomial.Point[] shares;
    private Polynomial.Point[] myshares;
    private BigInteger[][] commitmentsArray;
    private BigInteger[] ys;
    final ComplainState[][] complainStates;
    private Set<Integer> QUAL;

    private BigInteger x;
    private BigInteger y;
    private BigInteger[] commitments;

    public DKG(int t, int n, BigInteger x, Random random, BigInteger p, BigInteger q, BigInteger g,Network network) {
        super(t, n, x, random, p, q, g);

        this.commitmentsArray = new BigInteger[n][t + 1];
        this.shares = new Polynomial.Point[n];
        this.myshares = new Polynomial.Point[n];
        this.handler = new Handler();
        this.user = network.connect(handler);
        this.complainStates = new ComplainState[n][n];

        for (int i = 0; i < n; i ++){
            for (int j = 0 ; j < n ; j ++)
                complainStates[i][j] = ComplainState.Non;
        }

    }


    private void stage1(){
        int i = user.getID();
        BigInteger[] commitments = super.getCommitments();
        System.arraycopy(commitments, 0, commitmentsArray[i - 1], 0, commitmentsArray[i - 1].length);

        CommitmentMessage commitment;
        for (int k = 1; k <= commitmentsArray[i - 1].length; k ++){
            commitment = CommitmentMessage.newBuilder()
                            .setK(k)
                            .setCommitment(ByteString.copyFrom(commitmentsArray[i - 1][k - 1].toByteArray()))
                            .build();
            user.broadcast(Mail.Type.COMMITMENT,commitment);
        }

        SecretMessage secret;
        for (int j = 1; j <= n ; j++ ){
            if(j != i){
                myshares[j - 1] = super.getShare(j);
                secret = SecretMessage.newBuilder()
                            .setSecret(myshares[j - 1].asMessage())
                            .build();
                user.send(j, Mail.Type.SECRET,secret);
            }
        }
        while (!isStage1Complete()){
            try {
                Thread.sleep(30);
            } catch (InterruptedException e) {
                // do nothing
            }
        }
    }

    private boolean isStage1Complete(){
        int id = user.getID();
        for (int j = 1 ; j <= n ; j++){
            if(id != j && shares[j - 1] == null)
                return false;
        }

        for (int i = 0; i < commitmentsArray.length; i++){
            for (int j = 0; j < commitmentsArray[i].length; j++){
                if(commitmentsArray[i][j] == null)
                    return false;
            }
        }
        return true;
    }

    private boolean isValidSecret(Polynomial.Point secret, int i){
        int j = secret.x.intValue();
        return zpstar.multiply(g,secret.y).equals(verify(j,commitmentsArray[i],zpstar));
    }

    private void stage2(){
        int j = user.getID();
        QUAL = new HashSet<Integer>();
        ComplaintMessage complaint;
        for (int i = 1; i <= n ; i++ ){
            if(isValidSecret(shares[i - 1],j)) {
                QUAL.add(i);
            }else{
                //message = new Message(Type.Complaint, j)
                complaint = ComplaintMessage.newBuilder()
                                .setId(i)
                                .build();
                user.broadcast(Mail.Type.COMPLAINT,complaint);
                complainStates[j - 1][i - 1] = ComplainState.Waiting;
            }
        }
        //sending y after Complaints
        YMessage yMessage = YMessage.newBuilder()
                .setY(ByteString.copyFrom(super.getY().toByteArray()))
                .build();
        user.broadcast(Mail.Type.Y,yMessage);

        while (!isStage2Complete()){
            try {
                Thread.sleep(30);
            } catch (InterruptedException e) {
                // do nothing
            }
        }
    }

    private boolean isStage2Complete() {
        int id = user.getID();
        for (int j = 1; j <= n ; j++) {
            if (id != j && ys[j - 1] == null)
                return false;
        }
        for (int i = 0; i < complainStates.length;i++){
            for (int j =0 ; j < complainStates[i].length;j++){
                switch (complainStates[i][j]){
                    case Waiting:
                        return false;
                    default:
                        break;
                }
            }
        }
        return true;
    }

    private void stage3(){
        for (int i = 0; i < complainStates.length;i++){
            ComplainState state = ComplainState.Non;
            for (int j = 0 ; j < complainStates[i].length;j++){
                switch (complainStates[i][j]){
                    case Disqualified:
                        state = ComplainState.Disqualified;
                        break;
                    case NonDisqualified:
                        if(state == ComplainState.Non){
                            state = ComplainState.NonDisqualified;
                        }
                        break;
                    default:
                        break;
                }
            }
            switch (state){
                case Disqualified:
                    if(QUAL.contains(i + 1)){
                        QUAL.remove(i + 1);
                    }
                    break;
                case NonDisqualified:
                    if (!QUAL.contains(i + 1)){
                        QUAL.add(i + 1);
                    }
                    break;
                default:
                    break;
            }
        }
    }

    private void stage4(){
        BigInteger y = zpstar.zero();
        for (int i : QUAL) {
            y = zpstar.add(y , ys[i - 1]);
        }
        this.commitments = new BigInteger[t];
        BigInteger commitment;
        for (int k = 1; k <= t ; k++){
            commitment = zpstar.zero();
            for (int i : QUAL) {
                commitment = zpstar.add(commitment,commitmentsArray[i - 1][k - 1]);
            }
            commitments[k - 1] = commitment;
        }

        int j = user.getID();
        BigInteger x = BigInteger.ZERO;
        for (int i : QUAL) {
            x = x.add(shares[i - 1].y);
        }
        this.x = x.mod(q);
    }

    @Override
    public BigInteger getY() {
        return y;
    }

    @Override
    public BigInteger[] getCommitments() {
        return Arrays.clone(commitments);
    }

    @Override
    protected Polynomial.Point getShare(int i) {
        return null;
    }

    @Override
    public void run() {
        stage1();
        stage2();
        stage3();
        stage4();
    }

    private enum ComplainState{
        Non, Waiting,Disqualified,NonDisqualified
    }

    private class Handler implements Network.MailHandler {

        final int id;

        private Handler() {
            this.id = user.getID();
        }

        void handelSecretMessage(Mail mail) throws InvalidProtocolBufferException {
            if(shares[mail.getSender() - 1] == null) {
                SecretMessage secretMessage = SecretMessage.parseFrom(mail.getMessage());
                Polynomial.Point secret = new Polynomial.Point(secretMessage.getSecret());
                if(mail.getIsPrivate()){
                    shares[mail.getSender() - 1] = secret;
                }else{
                    int i = mail.getSender();
                    int j = secret.x.intValue();
                    switch (complainStates[i][j]){
                        case Waiting:
                            if(isValidSecret(secret,i)){
                                complainStates[i][j] = ComplainState.NonDisqualified;
                            }else{
                                complainStates[i][j] = ComplainState.Disqualified;
                            }
                            break;
                        default:
                            break;
                    }
                }
            }
        }

        void handelCommitmentMessage(Mail mail) throws InvalidProtocolBufferException {
            if(!mail.getIsPrivate()) { //broadcast only
                CommitmentMessage commitmentMessage = CommitmentMessage.parseFrom(mail.getMessage());
                if (commitmentsArray[mail.getSender() - 1][commitmentMessage.getK()] == null) {
                    BigInteger commitment = new BigInteger(commitmentMessage.getCommitment().toByteArray());
                    commitmentsArray[mail.getSender() - 1][commitmentMessage.getK()] = commitment;
                }
            }
        }

        void handelYMessage(Mail mail) throws InvalidProtocolBufferException {
            if(!mail.getIsPrivate()) { //broadcast only
                if (ys[mail.getSender() - 1] == null) {
                    YMessage yMessage = YMessage.parseFrom(mail.getMessage());
                    BigInteger y = new BigInteger(yMessage.getY().toByteArray());
                    ys[mail.getSender() - 1] = y;
                }
            }
        }

        void handelComplaintMessage(Mail mail) throws InvalidProtocolBufferException {
            if(!mail.getIsPrivate()) { //broadcast only
                ComplaintMessage complaintMessage = ComplaintMessage.parseFrom(mail.getMessage());
                int i = complaintMessage.getId();
                int j = mail.getSender();
                if(i == id){
                    user.broadcast(Mail.Type.SECRET,SecretMessage.newBuilder()
                            .setSecret(myshares[j].asMessage())
                            .build());
                }else{
                    switch (complainStates[i - 1][j - 1]){
                        case Non:
                            complainStates[i - 1][j - 1] = ComplainState.Waiting;
                            break;
                        default:
                            break;
                    }
                }
            }
        }

        @Override
        public void handel(Mail mail) throws InvalidProtocolBufferException {
           switch (mail.getType()){
               case SECRET:
                   handelSecretMessage(mail);
                   break;
               case COMMITMENT:
                   handelCommitmentMessage(mail);
                   break;
               case Y:
                   handelYMessage(mail);
                   break;
               case COMPLAINT:
                   handelComplaintMessage(mail);
                   break;
               default:
                   break;
           }
        }
    }
}
